import unittest
import sys
import os
from datetime import date, timedelta

from almapiwrapper.config import RecSet, NewLogicalSet, NewItemizedSet, Job, Reminder, fetch_reminders, fetch_libraries, Library, Location
from almapiwrapper.record import JsonData, XmlData
from almapiwrapper import config_log

config_log("test.log")
if os.getcwd().endswith('test'):
    os.chdir('..')

class TestReminder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        reminders = fetch_reminders(zone='NZ', env='S', entity_id='991170687152205501')
        if len(reminders) > 0:
            for r in reminders:
                r.delete()

        data = JsonData(filepath='test/data/reminder_test1.json')

        r = Reminder(zone='NZ', env='S', data=data, create_reminder=True)

    def test_get_reminder(self):
        reminders = fetch_reminders(zone='NZ', env='S', entity_id='991170687152205501')
        self.assertEqual(len(reminders), 1, 'Not able to fetch the reminder')
        self.assertEqual(reminders[0].entity_id, '991170687152205501', 'Reminder linked to the wrong record')
        self.assertEqual(reminders[0].reminder_type, 'Additional edition',
                         'Bad reminder type, should be "Additional edition"')
        self.assertEqual(reminders[0].status, 'NEW', 'Bad reminder status, should be "NEW"')


    def test_delete_reminder(self):
        reminders = fetch_reminders(zone='NZ', env='S', entity_id='991170687152205501')
        self.assertEqual(len(reminders), 1, 'Not able to fetch the reminder')
        self.assertEqual(reminders[0].entity_id, '991170687152205501', 'Reminder linked to the wrong record')
        reminders[0].delete()
        reminders = fetch_reminders(zone='NZ', env='S', entity_id='991170687152205501')
        self.assertEqual(len(reminders), 0, 'Not able to delete the reminder')

        data = JsonData(filepath='test/data/reminder_test1.json')
        Reminder(zone='NZ', env='S', data=data, create_reminder=True)

    def test_update_reminder(self):
        reminders = fetch_reminders(zone='NZ', env='S', entity_id='991170687152205501')
        self.assertEqual(len(reminders), 1, 'Not able to fetch the reminder')
        self.assertEqual(reminders[0].entity_id, '991170687152205501', 'Reminder linked to the wrong record')
        reminders[0].data['text'] = 'Test update'
        reminders[0].update()

        reminders = fetch_reminders(zone='NZ', env='S', entity_id='991170687152205501')
        self.assertEqual(len(reminders), 1, 'Not able to fetch the reminder')
        self.assertEqual(reminders[0].entity_id, '991170687152205501', 'Reminder linked to the wrong record')
        self.assertEqual(reminders[0].data['text'], 'Test update', 'Reminder not updated')


class TestRecSet(unittest.TestCase):

    def test_fetch_members_of_user_set(self):

        # Fetch set data
        s = RecSet('17866160420005504', 'UBS', 'S')
        members = s.get_members()

        self.assertEqual(len(members), s.get_members_number(), 'Not able to fetch all members')
        self.assertEqual(s.get_content_type(), 'USER', f'Bad type of content, should be "USER" '
                                                       f'and is "{s.get_content_type()}"')

    def test_fetch_members_of_bib_set(self):

        # Fetch set data (IEP is related to physical titles)
        s = RecSet('8020260160005504', 'UBS', 'S')
        members = s.get_members()

        self.assertEqual(len(members), s.get_members_number(), 'Not able to fetch all members')
        self.assertEqual(s.get_content_type(), 'IEP', f'Bad type of content, should be "IEP" '
                                                      f'and is "{s.get_content_type()}"')

    @unittest.skip
    def test_run_job(self):

        parameters = XmlData(filepath='test/data/config_job_parameters1.xml')

        job = Job('44', 'UBS', 'S')
        result = job.run(parameters)
        self.assertIsNotNone(result, "No return value when starting the job")
        self.assertFalse(job.error, 'The job cannot be started')
        self.assertIsNotNone(job.instance_id, "A job ID doesn't exist")

        instance = job.get_instance_info()
        self.assertIsNotNone(instance, "A job instance cannot be retrieved")

    def test_monitor_job(self):
        job = Job('44', 'UBS', 'S')

        instance = job.get_instance_info('17797121220005504')
        self.assertFalse(job.error, 'Not able to fetch a job instance')

        state = job.check_instance_state('17797121220005504')
        self.assertEqual(state['progress'], 100, 'Impossible to get instance info')
        self.assertEqual(state['status'], 'COMPLETED_SUCCESS', 'Impossible to get instance info')

        instances = job.get_instances()
        self.assertTrue(instances.content['total_record_count'] > 20)


class TestLibrary(unittest.TestCase):

    def test_fetch_libraries(self):
        libraries = fetch_libraries(zone='UBS', env='S')
        self.assertTrue(len(libraries) > 10, 'No libraries found')
        self.assertEqual(libraries[20].data['code'], libraries[20].code, 'Library object corrupted')

    def test_get_library(self):
        lib = Library('A100', 'UBS', 'S')
        self.assertEqual(lib.data['code'], 'A100', 'Library object corrupted')
        self.assertEqual(lib.data['name'], 'Basel - UB Hauptbibliothek', 'Library name not correct')


class TestLocation(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        loc = Location('UBS', 'A100', '100FH', 'S')
        if loc.fulfillment_unit != 'IZ_Glob_OS':
            loc.fulfillment_unit = 'IZ_Glob_OS'
            loc.update()

    def test_fetch_locations(self):
        lib = Library('A100', 'UBS', 'S')
        locations = lib.locations
        self.assertTrue(len(locations) > 10, 'No locations found')
        self.assertEqual(locations[20].data['code'], locations[20].code, 'Location object corrupted')

    def test_get_location(self):
        loc1 = Location('UBS', 'A100', '100FH', 'S')
        self.assertEqual(loc1.data['code'], '100FH', 'Location object corrupted')
        self.assertEqual(loc1.fulfillment_unit, 'IZ_Glob_OS', 'fulfillment unit should be "IZ_Glob_OS"')
        loc1.fulfillment_unit = 'IZ_ClosLib'
        loc1.update()
        loc2 = Location('UBS', 'A100', '100FH', 'S')
        self.assertEqual(loc2.fulfillment_unit, 'IZ_ClosLib', 'fulfillment unit should be "IZ_ClosLib"')

    def test_create_location(self):
        loc1 = Location('UBS', 'A100', '100FH', 'S')
        loc1.data['name'] = 'new test loc A100'
        loc1.data['external_name'] = 'new test loc A100'
        loc1.data['code'] = 'newlocA100'

        loc2 = Location('UBS', 'A100', env='S', data=loc1.data)
        _ = loc2.data

        loc2.create()

        loc3 = Location('UBS', 'A100', code='newlocA100', env='S')
        _ = loc3.data

        self.assertFalse(loc2.error, 'error when creating new location')
        self.assertFalse(loc3.error, 'error when fetching new location data')

        loc3.delete()
        loc3 = Location('UBS', 'A100', code='newlocA100', env='S')
        _ = loc3.data
        self.assertTrue(loc3.error, 'location should not exist anymore')

    @classmethod
    def tearDownClass(cls):
        loc = Location('UBS', 'A100', '100FH', 'S')
        loc.fulfillment_unit = 'IZ_Glob_OS'
        loc.update()


class TestOpenHours(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass
        # loc = Location('UBS', 'A100', '100FH', 'S')
        # cls.oh = loc.open_hours

    def test_get_open_hours(self):
        lib = Library('A100', 'UBS', 'S')

        self.assertTrue(len(lib.open_hours.data['open_hour']) > 0, 'No open hours found')
        self.assertEqual(lib.open_hours.data['open_hour'][-1]['desc'], 'Silvester', 'Sylvester must be last exception')

    def test_update_open_hours(self):
        new_open_hours = {
            "type": {
                "value": "EXCEPTION",
                "desc": "Exception"
            },
            "inherited": False,
            "desc": "test_exception",
            "from_date": f"{date.today()}Z",
            "to_date": f"{date.today() + timedelta(days=3)}Z",
            "from_hour": "00:00",
            "to_hour": "23:59",
            "status": {
                "value": "CLOSE",
                "desc": "Closed"
            }
        }
        lib1 = Library('A100', 'UBS', 'S')
        lib1.open_hours.data['open_hour'].append(new_open_hours)
        lib1.open_hours.update()
        self.assertEqual(lib1.open_hours.data['open_hour'][-1]['desc'],
                         'test_exception',
                         'Open hours object not updated')

        lib2 = Library('A100', 'UBS', 'S')
        is_test_exception = False
        for oh in lib2.open_hours.data['open_hour']:
            if oh['desc'] == 'test_exception':
                lib2.open_hours.data['open_hour'].remove(oh)
                is_test_exception = True

        self.assertTrue(is_test_exception, 'Test exception not found')

        lib2.open_hours.update()

        lib3 = Library('A100', 'UBS', 'S')
        is_test_exception = False
        for oh in lib3.open_hours.data['open_hour']:
            if oh['desc'] == 'test_exception':
                is_test_exception = True
        self.assertFalse(is_test_exception, 'Test exception not deleted')

    @classmethod
    def tearDownClass(cls):
        pass


if __name__ == '__main__':
    unittest.main()



from ..record import Record, check_error, JsonData
from typing import Optional, ClassVar, Literal
import logging
import almapiwrapper.users as userslib
from almapiwrapper.inventory import Item
from copy import deepcopy

class Request(Record):
    """Class representing a Users request

    :cvar api_base_url: url of the user api
    :ivar request_id: id of the request
    :ivar zone: initial value: zone of the request
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar user: :class:`almapiwrapper.users.User` either primary_id of the user or the user itself must be provided
    :ivar data: :class:`almapiwrapper.record.JsonData` with request data
    """

    api_base_url: ClassVar[str] = 'https://api-eu.hosted.exlibrisgroup.com/almaws/v1'

    def __init__(self,
                 request_id: Optional[str] = None,
                 user_id: Optional[str] = None,
                 zone: Optional[str] = None,
                 env: Optional[Literal['P', 'S']] = 'P',
                 user: Optional[userslib.User] = None,
                 data: Optional[JsonData] = None):
        """Constructor of Request Object

        """
        # Set user_id if provided
        self._user_id = user_id

        # Fetch env and zone from user if available
        if user is not None:
            zone = user.zone
            env = user.env

        super().__init__(zone, env, data)
        self.area = 'Users'
        self.format = 'json'

        self._request_id = request_id

        if user is not None:
            self._user_id = user.primary_id
            self.user = deepcopy(user)
        elif user_id is not None:
            self._user_id = user_id
            self.user = userslib.User(self._user_id, self.zone, self.env)
        elif data is not None and 'user_primary_id' in data.content:
            self._user_id = data.content['user_primary_id']
            self.user = userslib.User(self._user_id, self.zone, self.env)
        elif request_id is not None:
            self._user_id = None
            self.user = None
        else:
            logging.error('Missing information to construct a Request')
            self.error = True

    def __repr__(self):
        """Get a string representation of the object. Useful for logs.
        :return: str
        """
        return f"{self.__class__.__name__}('{self.request_id}', '{self.user.primary_id}', '{self.zone}', '{self.env}')"

    @property
    def request_id(self) -> str:
        """Property returning the loan ID

        It fetches it in a private attribute if not available
        in the data property.
        """
        if self._data is not None and 'request_id' in self.data:
            return self.data['request_id']
        else:
            return self._request_id

    def _fetch_data(self) -> Optional[JsonData]:
        """Fetch the json data of the loan

        :return: :class:`almapiwrapper.record.JsonData`"""
        r = self._api_call('get',
                           f'{self.api_base_url}/users/'
                           f'{self.user.primary_id if self._user_id is not None else "ALL"}/requests/'
                           f'{self.request_id}',
                           headers=self._get_headers())
        if r.ok is True:

            request_data = r.json()
            if self._user_id is None:
                self._user_id = request_data['user_primary_id']
                self.user = userslib.User(self._user_id, self.zone, self.env)
            logging.info(f'{repr(self)}: request data available')
            return JsonData(request_data)

        else:
            self._handle_error(r, f'{repr(self)}: unable to fetch user request')

    @check_error
    def create(self) -> 'userslib.Request':
        """create(self) -> 'userslib.Request'
        Create a new request

        Requests are created only at title level. MMS ID is required in request data
        to create a request.

        :return: object :class:`almapiwrapper.users.Request`"""

        # Decides if request is on item or on title level
        if 'item_id' in self.data and len(self.data['item_id']) > 0:
            params = {'item_pid': self.data['item_id']}
        else:
            params = {'mms_id': self.data['mms_id']}

        r = self._api_call('post',
                           f'{self.api_base_url}/users/{self.user.primary_id}/requests',
                           params=params,
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok is True:
            self.data = JsonData(r.json())
            logging.info(f'{repr(self)}: request created')
        else:
            self._handle_error(r, f'{repr(self)}: unable to create request')
        return self

    # @check_error
    # def _fetch_item(self) -> Item:
    #     """Fetch the loaned item
    #
    #     :return: :class:`almapiwrapper.inventory.Item`"""
    #     return Item(self.data['mms_id'],
    #                 self.data['holding_id'],
    #                 self.data['item_id'],
    #                 self.zone,
    #                 self.env)
    #
    # @check_error
    # def return_loan(self):
    #     pass
    #
    # def renew_loan(self) -> 'userslib.Loan':
    #     """Renew the loan
    #
    #     :return: object :class:`almapiwrapper.users.Loan`"""
    #     params = {'op': 'renew'}
    #     r = self._api_call('post',
    #                        f'{self.api_base_url_users}/{self.user.primary_id}/loans/{self.loan_id}',
    #                        headers=self._get_headers(),
    #                        data='{}',
    #                        params=params)
    #
    #     if r.ok is True:
    #         self.data = JsonData(r.json())
    #         if self.data['last_renew_status']['desc'] == 'Renewed Successfully':
    #             logging.info(f'{repr(self)}: loan renewed => new due date: {self.data["due_date"]}')
    #         else:
    #             logging.warning(f'{repr(self)}: loan not renewed: {self.data["last_renew_status"]["desc"]}')
    #     else:
    #         self._handle_error(r, f'{repr(self)}: unable to renew loan')
    #     return self
    #
    # @check_error
    # def change_due_date(self, new_due_date: str) -> 'userslib.Loan':
    #     """Change due date of the loan
    #
    #     :param new_due_date: str : new due date in format YYYY-MM-DD
    #
    #     :return: object :class:`almapiwrapper.users.Loan`"""
    #     # self.data['due_date'] = new_due_date
    #     r = self._api_call('put',
    #                        f'{self.api_base_url_users}/{self.user.primary_id}/loans/{self.loan_id}',
    #                        headers=self._get_headers(),
    #                        data=bytes(JsonData(content={'due_date': new_due_date})))
    #
    #     if r.ok is True:
    #         self.data = JsonData(r.json())
    #         logging.info(f'{repr(self)}: due date changed to {new_due_date}')
    #     else:
    #         self._handle_error(r, f'{repr(self)}: unable to change due date of the loan')
    #     return self
    #
    # @check_error
    # def return_loan(self):
    #     pass

    @check_error
    def cancel(self, reason: Optional[str] = 'CannotBeFulfilled',
               note: Optional[str] = None,
               notify_user: Optional[bool]=True) -> None:
        """cancel(self, reason: Optional[str] = 'Cancelled by patron', note: Optional[str] = None, notify_user: Optional[bool]=True) -> None
        Cancel the request

        :param reason: str : reason for the cancellation
        :param note: str : note for the cancellation
        :param notify_user: bool : notify the user about the cancellation

        :return: None"""
        params = {'reason': reason,
                  'notify_user': notify_user}
        if note is not None:
            params['note'] = note
        r = self._api_call('delete',
                           f'{self.api_base_url}/users/{self.user.primary_id}/requests/{self.request_id}',
                           headers=self._get_headers(),
                           params=params)

        if r.ok is True:
            logging.info(f'{repr(self)}: request cancelled')
        else:
            self._handle_error(r, f'{repr(self)}: unable to cancel request')

    @check_error
    def save(self) -> 'userslib.Request':
        """Save a user loan record in the 'records' folder

        When saved, a suffix is added to the file path with the version.
        Example: records/<primary_id>/request_<IZ>_<request_id>_<version>.xml

        :return: object :class:`almapiwrapper.users.Request`
        """
        filepath = f'records/{self.user.primary_id}/request_{self.zone}_{self.request_id}.json'
        self._save_from_path(filepath)
        return self



"""This module allow to create and manage sets in Alma"""
from typing import Literal, Optional, Union, List
import logging
from ..record import XmlData, Record, check_error
from ..inventory import IzBib, NzBib
from ..users import User
from lxml import etree


class RecSet(Record):
    """RecSet(self, set_id: Optional[str] = None, zone: Optional[str] = None, env: Literal['P', 'S'] = 'P', name: Optional[str] = None, data: Optional[XmlData] = None) -> None

    Class representing a set of records

    To get a set it is possible to provide either the set id or the name of the set.

    :param set_id: ID of the set
    :param zone: zone of the set
    :param env: environment of the entity: 'P' for production and 'S' for sandbox
    :param name: name of the set, can be used instead of the set_id
    :param data: :class:`almapiwrapper.record.XmlData` with the data describing the set. If provided, the set_id and name are not required.

    :ivar set_id: initial value: ID of the set
    :ivar zone: initial value: zone of the set
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar name: initial value: name of the set
    :ivar data: initial value: data describing the set
    """

    def __new__(cls, *args, **kwargs) -> Union['ItemizedSet', 'LogicalSet', 'RecSet']:
        """Constructor of `RecSet`"""
        if 'data' not in kwargs:
            return super().__new__(cls)
        data = kwargs['data'].content
        if data.find('.//type').text == 'ITEMIZED':
            return super().__new__(ItemizedSet)

        return super().__new__(LogicalSet)

    def __init__(self,
                 set_id: Optional[str] = None,
                 zone: Optional[str] = None,
                 env: Literal['P', 'S'] = 'P',
                 name: Optional[str] = None,
                 data: Optional[XmlData] = None) -> None:
        """Constructor of `RecSet`
        """
        super().__init__(zone, env, data)
        self.area = 'Conf'
        self.format = 'xml'
        self._name = name
        if data is not None and set_id is None:
            set_id = data.content.find('.//id').text
        self._set_id = set_id
        self._members = None

    def _fetch_data(self) -> Optional[XmlData]:
        """This method fetch the data describing the set.

        It is possible to fetch the data of a set by providing the set_id or the name of the set.

        :return: :class:`almapiwrapper.record.XmlData`
        """
        # set_id is not provided. We try to fetch it from the name.
        if self._set_id is None and self._name is not None:
            r = self._api_call('get',
                               f'{self.api_base_url}/conf/sets',
                               params={'q': f'name~{self._name}'},
                               headers=self._get_headers())
            if r.ok is True and XmlData(r.content).content.find('.//set') is not None:
                logging.info(f'{repr(self)}: set data available')
                set_data = XmlData(r.content).content.find('.//set')

                # We store the set_id for fetching the complete data
                self._set_id = set_data.find('.//id').text

            else:
                logging.error(f'{repr(self)}: unable to fetch set data')
                self.error = True
                return None

        # set_id is provided. We fetch the data.
        r = self._api_call('get',
                           f'{self.api_base_url}/conf/sets/{self.set_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: set data available')
            return XmlData(r.content)
        else:
            self._handle_error(r, 'unable to fetch set data')

    @property
    def name(self) -> str:
        """Property returning the name of the set.

        :return: str containing set name
        """
        if self._name is None:
            name = self.data.find('.//name')
            if name is not None:
                self._name = name.text

        return self._name

    @property
    def set_id(self) -> str:
        """Property returning the name of the set.

        :return: str containing set name
        """
        if self._set_id is None:
            set_id = self.data.find('.//id')
            if set_id is not None:
                self._set_id = set_id.text

        return self._set_id

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        if self._set_id is not None:
            return f"{self.__class__.__name__}('{self.set_id}', '{self.zone}', '{self.env}')"
        else:
            return f"{self.__class__.__name__}(name='{self.name}', zone='{self.zone}', env='{self.env}')"

    @check_error
    def save(self) -> Union['RecSet', 'ItemizedSet', 'LogicalSet']:
        """save() -> Union['RecSet', 'ItemizedSet', 'LogicalSet']
        Save a set record in the 'records' folder

        When saved, a suffix is added to the file path with the version.
        Example: records/set_<IZ>_<set_id>_<version>.xml

        :return: object :class:`almapiwrapper.users.Fee`

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        filepath = f'records/recset/set_{self.zone}_{self.name}.xml'
        self._save_from_path(filepath)
        return self

    @check_error
    def get_members_number(self) -> int:
        """get_members_number() -> int
        Return the number of members

        :return: int with the number of records

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        # If number of records is not provided, it's a new set and data need to be refreshed
        if self.data.find('.//number_of_members') is None:
            self.data = self._fetch_data()
        return int(self.data.find('.//number_of_members').text)

    @check_error
    def get_set_type(self) -> str:
        """get_set_type() -> str
        Return the type of the set

        :return: str indicating the type of the set

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        return self.data.find('.//type').text

    @check_error
    def get_content_type(self) -> str:
        """get_content_type() -> str
        Return the type of the set

        :return: str indicating the type of the set

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        return self.data.find('.//content').text

    @check_error
    def update(self) -> Union['RecSet', 'ItemizedSet', 'LogicalSet']:
        """update() -> Union['RecSet', 'ItemizedSet', 'LogicalSet']
        Update set data.

        :return: the updated set object

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        r = self._api_call('put',
                           f'{self.api_base_url}/conf/sets/{self.set_id}',
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok is True:
            self.data = XmlData(r.content)
            logging.info(f'{repr(self)}: set data updated')
        else:
            self._handle_error(r, 'unable to update set data')

        return self


    @check_error
    def delete(self) -> None:
        """delete() -> None
        Delete a set

        :return: None

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """

        r = self._api_call('delete',
                           f'{self.api_base_url}/conf/sets/{self.set_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: set deleted')
            return
        else:
            print(r.text)
            self._handle_error(r, f'unable to delete the set')

    @check_error
    def get_members(self) -> list[Record]:
        """get_members() -> list[Record]
        Return a list with all the records of the set

        :return: list of records depending on the type of the set

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """

        # Prevent to fetch data twice, if already available
        if self._members is not None:
            members = self._members
            return members
        else:
            self._members = []
            while len(self._members) < self.get_members_number():
                r = self._api_call('get',
                                   f'{self.api_base_url}/conf/sets/{self.set_id}/members',
                                   params={'limit': '100', 'offset': str(len(self._members))},
                                   headers=self._get_headers())
                if r.ok is False:
                    self._handle_error(r, f'{repr(self)}: unable to fetch set members')
                    return self._members
                else:
                    rec_ids = [rec_id.text for rec_id in XmlData(r.content).content.findall('.//member/id')]

                    # Build members -> IZ mms_id
                    if self.get_content_type() in ['BIB_MMS', 'IEP'] and self.zone != 'NZ':
                        self._members += [IzBib(rec_id, self.zone, self.env) for rec_id in rec_ids]

                    # Build members -> IZ mms_id
                    elif self.get_content_type() in ['BIB_MMS'] and self.zone == 'NZ':
                        self._members += [NzBib(rec_id, self.env) for rec_id in rec_ids]

                    # Build members -> users
                    elif self.get_content_type() == 'USER':
                        self._members += [User(rec_id, self.zone, self.env) for rec_id in rec_ids]

                    # Type of content not integrated
                    else:
                        self._members += rec_ids
                        logging.warning(f'{repr(self)}: "{self.get_content_type()}" not integrated, only id are returned.')

                    logging.info(f'{repr(self)}: {len(self._members)} / {self.get_members_number()} members fetched')

            return self._members


class LogicalSet(RecSet):
    """LogicalSet(set_id: Optional[str] = None, zone: Optional[str] = None, env: Literal['P', 'S'] = 'P', name: Optional[str] = None, data: Optional[XmlData] = None) -> None

    Class representing a logical set of records

    :param set_id: ID of the set
    :param zone: zone of the set
    :param env: environment of the entity: 'P' for production and 'S' for sandbox
    :param name: name of the set, can be used instead of the set_id
    :param data: :class:`almapiwrapper.record.XmlData` with the data describing the set. If provided, the set_id and name are not required.


    :ivar set_id: initial value: ID of the set
    :ivar zone: initial value: zone of the set
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar name: initial value: name of the set
    :ivar data: initial value: data describing the set
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

class ItemizedSet(RecSet):
    """ItemizedSet(set_id: Optional[str] = None, zone: Optional[str] = None, env: Literal['P', 'S'] = 'P', name: Optional[str] = None, data: Optional[XmlData] = None) -> None

    Class representing an itemized set of records

    :param set_id: ID of the set
    :param zone: zone of the set
    :param env: environment of the entity: 'P' for production and 'S' for sandbox
    :param name: name of the set, can be used instead of the set_id
    :param data: :class:`almapiwrapper.record.XmlData` with the data describing the set. If provided, the set_id and name are not required.

    :ivar set_id: initial value: ID of the set
    :ivar zone: initial value: zone of the set
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar name: initial value: name of the set
    :ivar data: initial value: data describing the set
    """
    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    @check_error
    def add_members(self,
                    record_ids: List[str],
                    fail_on_invalid_id: Optional[bool] = True) -> 'ItemizedSet':
        """add_members(record_ids: List[str], fail_on_invalid_id: Optional[bool] = True) -> 'ItemizedSet'
        Add members to a set

        :param record_ids: list of record ids to add to the set
        :param fail_on_invalid_id: if True, raise an error if one of the record id is invalid
        :return: `ItemizedSet` object

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        # params of the add members API call
        params = {'op': 'add_members',
                  'fail_on_invalid_id': fail_on_invalid_id}

        # Split the list of record ids in chunks of 1000
        for i in range(len(record_ids) // 1000 + 1):

            # Remove the members node if existing, useful when more than 1000 members
            old_members = self.data.find('members')
            if old_members is not None:
                self.data.remove(old_members)

            members = etree.Element('members')
            for record_id in record_ids[i*1000:(i+1)*1000]:
                etree.SubElement(etree.SubElement(members, 'member'), 'id').text = record_id

            self.data.append(members)

            number_of_members_before_update = self.get_members_number()

            r = self._api_call('post',
                               f'{self.api_base_url}/conf/sets/{self.set_id}',
                               headers=self._get_headers(),
                               params=params,
                               data=bytes(self))

            if r.ok is True:
                self.data = XmlData(r.content)
                number_of_members_after_update = self.get_members_number()
                self._members = None
                logging.info(f'{repr(self)}: adding members succeeded, '
                             f'{number_of_members_after_update - number_of_members_before_update} members added '
                             f'({number_of_members_after_update} members in total)')
            else:
                self._handle_error(r, 'unable to add members to the set')
                break

        return self

class NewLogicalSet(RecSet):
    """NewLogicalSet(zone: str, env: Literal['P', 'S'] = 'P', name: Optional[str] = None, description: Optional[str] = None, query: Optional[str] = None, created_by: Optional[str] = None, private: Optional[bool] = False, data: Optional[XmlData] = None) -> None
    Class representing a new logical set of records

    To create a new set, it is possible to provide directly the XML data or
    to give all the required parameters.

    :param name: name of the set
    :param zone: zone of the set
    :param env: environment of the set: 'P' for production and 'S' for sandbox
    :param description: description of the set
    :param query: query to build the set
    :param created_by: user who created the set
    :param private: set as private or not
    :param data: :class:`almapiwrapper.record.XmlData` object, can be used instead of the XML data of the set

    :ivar zone: initial value: zone of the set
    :ivar env: initial value: environment of the set: 'P' for production and 'S' for sandbox
    :ivar data: initial value: :class:`almapiwrapper.record.XmlData`
    :ivar area: initial value: area of the set: will be "Conf"
    :ivar format: initial value: data format of the set: will be "xml"
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)
    def __init__(self,
                 zone: str,
                 env: Literal['P', 'S'] = 'P',
                 name: Optional[str] = None,
                 description: Optional[str] = None,
                 query: Optional[str] = None,
                 created_by: Optional[str] = None,
                 private: Optional[bool] = False,
                 data: Optional[XmlData] = None) -> None:
        """
        Constructor of `NewLogicalSet`
        """
        super(RecSet, self).__init__(zone, env, data)
        self.area = 'Conf'
        self.format = 'xml'
        if self._data is None and name is not None and query is not None and created_by is not None:
            self._data = self._create_data(name, query, description, created_by, private)
        if self._data is None:
            logging.error('NewLogicalSet: missing parameters to create a new set')

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.zone}', '{self.env}', '{self.data.find('name').text}', " \
               f"'{self.data.find('description').text}', '{self.data.find('query').text}', " \
               f"'{self.data.find('created_by').text}', {self.data.find('private').text=='true'})"

    @staticmethod
    def _create_data(name, query, description, created_by, private) -> XmlData:
        """Create the XML data for a new set

        :param name: name of the set
        :param query: query of the set
        :param description: description of the set
        :param created_by: user who created the set
        :param private: bool indicating if set is private or public
        :return: XML data
        """
        content = query.split()[0]
        root = etree.Element('set')
        etree.SubElement(root, 'name').text = name
        etree.SubElement(root, 'type').text = 'LOGICAL'
        etree.SubElement(root, 'content').text = content
        etree.SubElement(root, 'description').text = description
        etree.SubElement(root, 'query').text = query
        etree.SubElement(root, 'created_by').text = created_by
        etree.SubElement(root, 'private').text = str(private).lower()
        return XmlData(etree.tostring(root))

    @check_error
    def create(self) -> Union['LogicalSet', 'NewLogicalSet']:
        """create() -> Union['LogicalSet', 'NewLogicalSet']
        Create the logical set

        :return: :class:`almapiwrapper.config.LogicalSet` or :class:`almapiwrapper.config.NewLogicalSet` in case of error

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        r = self._api_call('post',
                           f'{self.api_base_url}/conf/sets',
                           headers=self._get_headers(),
                           data=bytes(self))
        if r.ok is True:
            self._data = XmlData(r.content)
            set_id = self.data.find('.//id').text
            new_rec = RecSet(set_id, self.zone, self.env, data=self._data)
            logging.info(f'{repr(new_rec)}: set created')
            return new_rec
        else:
            self._handle_error(r, f'unable to create the set')
            return self


class NewItemizedSet(RecSet):
    """NewItemizedSet(zone: str, env: Literal['P', 'S'] = 'P', name: Optional[str] = None, content: Optional[str] = None, description: Optional[str] = None, private: Optional[bool] = False, members: Optional[List[str]] = None, from_logical_set: Optional[Union[str, `LogicalSet`]] = None, data: Optional[XmlData] = None) -> None
    Class representing a new itemized set of records

    To create a new set, it is possible to provide directly the XML data or
    to give all the required parameters.

    :param zone: zone of the set
    :param env: environment of the set: 'P' for production and 'S' for sandbox
    :param name: name of the set
    :param content: content of the set for example 'BIB_MMS' or 'USER'
    :param description: description of the set
    :param query: query to build the set
    :param private: set as private or not
    :param members: list of members of the set, must be a list of primary ids
    :param from_logical_set: set_id or :class:`almapiwrapper.config.LogicalSet` object to create the set from
    :param data: :class:`almapiwrapper.record.XmlData` object, can be used instead of the XML data of the set

    :ivar zone: initial value: zone of the set
    :ivar env: initial value: environment of the set: 'P' for production and 'S' for sandbox
    :ivar data: initial value: :class:`almapiwrapper.record.XmlData`
    :ivar from_logical_set: initial value: set_id string of `LogicalSet` to create the `ItemizedSet` from
    :ivar area: initial value: area of the set: will be "Conf"
    :ivar members: initial value: list of members of the set, must be a list of primary ids
    :ivar format: initial value: data format of the set: will be "xml"
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)
    def __init__(self,
                 zone: str,
                 env: Literal['P', 'S'] = 'P',
                 name: Optional[str] = None,
                 content: Optional[str] = None,
                 description: Optional[str] = None,
                 private: Optional[bool] = False,
                 members: Optional[List[str]] = None,
                 from_logical_set: Optional[Union[str, RecSet]] = None,
                 data: Optional[XmlData] = None) -> None:
        """Constructor of `NewItemizedSet`
        """
        super(RecSet, self).__init__(zone, env, data)
        self.area = 'Conf'
        self.format = 'xml'
        self.members = members

        # Will be None if itemized set is not created from a logical set. Set can be indicated by its ID or by a RecSet
        self.from_logical_set = from_logical_set.set_id if isinstance(from_logical_set, LogicalSet) is True or \
            isinstance(from_logical_set, RecSet) is True else from_logical_set

        # If itemized set is created from a logical set, we can get the name and the content from the logical set
        if  self.from_logical_set is not None and (name is None or content is None):
            if isinstance(from_logical_set, str) is True:
                from_logical_set = LogicalSet(from_logical_set, zone, env)
            if name is None:
                name = from_logical_set.data.find('name').text + '_itemized'
            if content is None:
                content = from_logical_set.data.find('content').text

        # Check if required parameters are provided
        if self._data is None and name is not None and content is not None:
            self._data = self._create_data(name, content, description, private, self.from_logical_set)
        else:
            logging.error('NewLogicalSet: missing parameters to create a new set')

    @staticmethod
    def _create_data(name: str,
                     content: str,
                     description: Optional[str] = None,
                     private: Optional[bool] = False,
                     from_logical_set: Optional[str] = None) -> XmlData:
        """Create the XML data for a new set

        :param name: name of the set
        :param content: content of the set
        :param description: description of the set
        :param private: bool indicating if set is private or public
        :return: :class:`almapiwrapper.record.XmlData` object

        """
        root = etree.Element('set')
        etree.SubElement(root, 'name').text = name
        etree.SubElement(root, 'type').text = 'ITEMIZED'
        etree.SubElement(root, 'content').text = content
        etree.SubElement(root, 'description').text = description if description is not None else ''
        etree.SubElement(root, 'private').text = str(private).lower()
        if from_logical_set is not None:
            etree.SubElement(root, 'fromLogicalSet').text = from_logical_set
        return XmlData(etree.tostring(root))

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.zone}', '{self.env}', '{self.data.find('name').text}', " \
               f"'{self.data.find('content').text}', " \
               f"'{self.data.find('description').text if self.data.find('description').text is not None else ''}', " \
               f"{self.data.find('private').text == 'true'})"

    @check_error
    def create(self) -> Union['ItemizedSet', 'NewItemizedSet'] :
        """create() -> Union['LogicalSet', 'NewLogicalSet']
        Create the itemized set

        :return: :class:`almapiwrapper.config.ItemizedSet` or :class:`almapiwrapper.config.NewItemizedSet` in case of error

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        params = {'from_logical_set': self.from_logical_set} if self.from_logical_set is not None else {}

        r = self._api_call('post',
                           f'{self.api_base_url}/conf/sets',
                           headers=self._get_headers(),
                           params=params,
                           data=bytes(self))
        if r.ok is True:
            logging.info(f'{repr(self)}: set created')
            self._data = XmlData(r.content)
            set_id = self.data.find('.//id').text
            new_rec = RecSet(set_id, self.zone, self.env, data=self._data)
            logging.info(f'{repr(new_rec)}: set created')
            if self.members is not None:
                new_rec.add_members(self.members)
            return new_rec
        else:
            self._handle_error(r, f'unable to create the set')
            return self



import almapiwrapper.users as userslib
from typing import List, Optional, Literal, Union
from ..apikeys import ApiKeys
from ..record import Record, JsonData
import requests
import logging
import time
from copy import deepcopy
import re
from json import JSONDecodeError
from lxml import etree


def fetch_users(q: str, zone: str, env: Optional[Literal['P', 'S']] = 'P') -> List[userslib.User]:
    """Search users in IZ according to a request

    :param q: request in API syntax. "user_group~06" will search every linked account
    :param zone: code of the IZ, NZ or "all" for all IZs NZ excepted.
    :param env: "P" for production, "S" for sandbox. Default is production
    :return: list of :class:`almapiwrapper.users.User`
    """
    error = False
    users = []
    list_iz = ApiKeys().get_iz_codes() if zone == 'all' else [zone]
    for z in list_iz:

        # Interrupt in case of any error
        if error is True:
            break

        # Init counters
        offset = 0
        nb_total_records = 0

        # Handle offset if more than 100 results are available
        while offset == 0 or offset < nb_total_records:

            # Make request
            r = requests.get(f'{userslib.User.api_base_url}',
                             params={'q': q, 'limit': 100, 'offset': offset},
                             headers=Record.build_headers(data_format='json', env=env,
                                                          zone=z, rights='RW', area='Users'))

            # Check result
            if r.ok is True:
                users_list = JsonData(r.json())
                nb_total_records = int(users_list.content['total_record_count'])

                if 'user' in users_list.content and users_list.content['user'] is not None:
                    users += [userslib.User(user['primary_id'], z, env) for user in users_list.content['user']]
                    nb_users = len(users_list.content['user'])
                else:
                    nb_users = 0

                logging.info(f'fetch_users("{q}", "{z}", "{env}"): '
                             f'{offset + nb_users} / '
                             f'{nb_total_records} users data available')
                offset += 100
            else:
                _handle_error(q, r, f'unable to fetch data', z, env)
                error = True
                break

    return users


def fetch_user_in_all_iz(primary_id: str, env: Optional[Literal['P', 'S']] = 'P') -> List[userslib.User]:
    """Fetch by primary ID a user in all IZ

    :param primary_id: primary ID of the user to search across all IZs
    :param env: "P" for production, "S" for sandbox. Default is production
    :return: list of :class:`almapiwrapper.users.User`
    """

    list_iz = ApiKeys().get_iz_codes(env=env)
    users = []
    for iz in list_iz:
        users_temp = fetch_users(f'primary_id~{primary_id.replace(" ", "+")}', iz, env)
        if len(users_temp) == 1:
            users += users_temp

    return users


def check_synchro(nz_users: Union[List[userslib.User], userslib.User]) -> Optional[List[userslib.User]]:
    """Test if a NZ user is synchronized with copies of the account across IZs.
    :param nz_users: list of :class:`almapiwrapper.users.User` or only one :class:`almapiwrapper.users.User`
    :return: list of :class:`almapiwrapper.users.User` with not synchronized IZ user accounts
    """

    not_synchro_iz_users = []
    if type(nz_users) is not list:
        nz_users = [nz_users]

    nb_users = len(nz_users)
    nz_users = [user for user in nz_users if user.zone == 'NZ']

    if len(nz_users) < nb_users:
        logging.error(f'Impossible to check synchronization on not NZ account')
        return None

    for nz_user in nz_users:
        nz_user.add_synchro_note()

    time.sleep(300)

    for nz_user in nz_users:
        nz_user.save()
        iz_users = fetch_user_in_all_iz(nz_user.primary_id, nz_user.env)

        for iz_user in iz_users:
            iz_user.save()
            if iz_user.check_synchro_note() is False:
                not_synchro_iz_users.append(iz_user)

    for nz_user in nz_users:
        nz_user.remove_synchro_note()

    if len(not_synchro_iz_users) > 0:
        logging.warning(f'Count of not synchronized users: {len(not_synchro_iz_users)}')
    else:
        logging.info(f'Count of not synchronized users: {len(not_synchro_iz_users)}')

    return not_synchro_iz_users

def force_synchro(nz_users: Union[List[userslib.User], userslib.User]) -> List[str]:
    """Force synchronization of a NZ user with copies of the account across IZs.
    :param nz_users: list of :class:`almapiwrapper.users.User` or only one :class:`almapiwrapper.users.User`
    :return: list of errors
    """

    error_msg = []

    if type(nz_users) is not list:
        nz_users = [nz_users]

    nb_users = len(nz_users)
    nz_users = [user for user in nz_users if user.zone == 'NZ']

    if len(nz_users) < nb_users:
        logging.error(f'Impossible to force synchronization on not NZ account')
        error_msg.append(f'Impossible to force synchronization on not NZ account')
        return error_msg

    for nz_user in nz_users:
        _ = nz_user.data
        if nz_user.error is True:
            error_msg.append(f'{repr(nz_user)}: no NZ account -> impossible to force synchronization')
            logging.error(f'{repr(nz_user)}: no NZ account -> impossible to force synchronization')

            continue
        iz_users = fetch_user_in_all_iz(nz_user.primary_id, nz_user.env)
        _ = iz_users

        for iz_user in iz_users:
            iz_user.save()

            # Copy contact_info
            iz_user.data['contact_info'] = deepcopy(nz_user.data['contact_info'])
            for segment in iz_user.data['contact_info']:
                for entry in iz_user.data['contact_info'][segment]:
                    entry['segment_type'] = 'External'

            # Copy identifiers
            nz_ids = [nz_id['value'] for nz_id in nz_user.data['user_identifier']]
            local_ids = [u_id for u_id in iz_user.data['user_identifier'] if
                         u_id['value'] not in nz_ids and u_id['id_type']['value'] == '03']

            # Transform segment of NZ ids to External before copying into IZ account
            ids_to_be_copied = deepcopy(nz_user.data['user_identifier'])
            for id_to_be_copied in ids_to_be_copied:
                id_to_be_copied['segment_type'] = 'External'
            iz_user.data['user_identifier'] = local_ids + ids_to_be_copied
    
            # Copy notes
            local_notes = [u_note for u_note in iz_user.data['user_note'] if u_note['segment_type'] == 'Internal']

            # Make nz notes as external when copied into IZ account
            notes_to_be_copied = deepcopy(nz_user.data['user_note'])
            for note in notes_to_be_copied:
                note['segment_type'] = 'External'

            iz_user.data['user_note'] = local_notes + notes_to_be_copied
            iz_user.dedup_notes()
    
            # Copy user_group
            nz_user_group = nz_user.data['user_group']['value']
            local_user_group = iz_user.data['user_group']['value']
            local_user_group_name = iz_user.data['user_group']['desc']

            if local_user_group != nz_user_group and 'MM-' not in local_user_group_name:
                    iz_user.data['user_group']['value'] = nz_user_group


            if iz_user.error is True:
                error_msg.append(f'{repr(iz_user)}: error on update / {iz_user.error_msg}')

            # Copy blocks
            local_blocks = [u_block for u_block in iz_user.data['user_block'] if u_block['segment_type'] == 'Internal']

            # Make nz blocks as external when copied into IZ account
            blocks_to_be_copied = deepcopy(nz_user.data['user_block'])
            for block in blocks_to_be_copied:
                block['segment_type'] = 'External'

            iz_user.data['user_block'] = local_blocks + blocks_to_be_copied
            iz_user.dedup_notes()

            # Copy first name and last name
            for field in ['last_name', 'first_name', 'middle_name']:
                if nz_user.data.get(field, '') != iz_user.data.get(field, ''):
                    iz_user.data[field] = nz_user.data[field]

            iz_user.update(override=['user_group'])

    return error_msg

def _handle_error(q: str, r: requests.models.Response, msg: str, zone: str, env: str):
    """Set the record error attribute to True and write the logs about the error

    :param r: request response of the api
    :param msg: context message of the error
    :return: None
    """
    try:
        json_data = r.json()
        error_message = json_data['errorList']['error'][0]['errorMessage']
    except JSONDecodeError:
        xml = etree.fromstring(r.content)
        error_message = xml.find('.//{http://com/exlibris/urm/general/xmlbeans}errorMessage').text

    logging.error(f'fetch_users("{q}", "{zone}", "{env}") - {r.status_code}: '
                  f'{msg} / {error_message}')


if __name__ == "__main__":
    pass



from ..record import Record, check_error, JsonData
from typing import Optional, ClassVar, Literal
import logging
import almapiwrapper.users as userslib


class Fee(Record):
    """Class representing a Users fee

    :cvar api_base_url_users: url of the user api
    :ivar fee_id: id of the fee
    :ivar zone: initial value: zone of the fee
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar user: :class:`almapiwrapper.users.User` either primary_id of the user or the user itself must be provided
    :ivar data: :class:`almapiwrapper.record.JsonData` with fee data"""

    api_base_url_users: ClassVar[str] = 'https://api-eu.hosted.exlibrisgroup.com/almaws/v1/users'

    def __init__(self,
                 fee_id: Optional[str] = None,
                 primary_id: Optional[str] = None,
                 zone: Optional[str] = None,
                 env: Optional[Literal['P', 'S']] = 'P',
                 user: Optional[userslib.User] = None,
                 data: Optional[JsonData] = None,
                 create_fee: Optional[bool] = False):
        """Constructor of Fee Object
        """
        # Fetch env and zone from user if available
        if user is not None:
            zone = user.zone
            env = user.env

        super().__init__(zone, env, data)
        self.area = 'Users'
        self.format = 'json'
        self._fee_id = fee_id
        if primary_id is not None:
            self.user = userslib.User(primary_id, self.zone, self.env)
        elif user is not None:
            self.user = user
        else:
            logging.error('Missing information to construct a Fee')
            self.error = True

        if create_fee is True:
            self._create_fee()

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.
        :return: str
        """
        return f"{self.__class__.__name__}('{self.fee_id}', '{self.user.primary_id}', '{self.zone}', '{self.env}')"

    def _fetch_data(self) -> Optional[JsonData]:
        """Fetch the json data of the fee

        :return: :class:`almapiwrapper.record.JsonData`"""
        r = self._api_call('get',
                           f'{self.api_base_url_users}/{self.user.primary_id}/fees/{self.fee_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: fee data available')
            fee_data = r.json()

            return JsonData(fee_data)

        else:
            self._handle_error(r, f'{repr(self)}: unable to fetch user fee')

    @property
    def fee_id(self) -> str:
        """Property returning the fee ID

        It fetches it in a private attribute if not available
        in the data property.
        """
        if self._data is not None:
            return self.data['id']
        else:
            return self._fee_id

    def _create_fee(self):
        """Create a fee to the user with the provided data.

        :return: object :class:`almapiwrapper.users.Fee`
        """
        r = self._api_call('post',
                           f'{self.api_base_url_users}/{self.user.primary_id}/fees',
                           headers=self._get_headers(),
                           data=bytes(self))
        if r.ok is True:
            self.data = JsonData(r.json())
            logging.info(f'{repr(self)}: fee created')
        else:
            self._handle_error(r, 'unable to create the new fee')

    @check_error
    def save(self) -> 'userslib.Fee':
        """Save a user fee record in the 'records' folder

        When saved, a suffix is added to the file path with the version.
        Example: records/<primary_id>/fee_<IZ>_<fee_id>_<version>.xml

        :return: object :class:`almapiwrapper.users.Fee`
        """
        filepath = f'records/{self.user.primary_id}/fee_{self.zone}_{self.fee_id}.json'
        self._save_from_path(filepath)
        return self

    @check_error
    def operate(self,
                op: Literal['pay', 'waive', 'dispute', 'restore'],
                amount: Optional[float] = None,
                comment: Optional[str] = None,
                method: Literal['CREDIT_CARD', 'ONLINE', 'CASH'] = None,
                reason: Optional[str] = None,
                external_transaction_id: Optional[str] = None):
        """Operate on a fee

        :param op: can be 'pay', 'waive', 'dispute', 'restore'
        :param amount: optional amount. For pay and waive, it will be the balance as default value
        :param comment: free text associated to the operation
        :param method: for pay or waive operations, can be: 'CREDIT_CARD', 'ONLINE', 'CASH'
        :param reason: a value of FineFeeTransactionReason table. Required only for waiving fee
        :param external_transaction_id: external id of the bursar system

        :return: object :class:`almapiwrapper.users.Fee`
        """
        # Per default pay the entire fee
        if op in ['pay', 'waive'] and amount is None:
            amount = self.data['balance']

        # Set a default payment method if required
        if op == 'pay' and method is None:
            method = 'ONLINE'

        # Set a default value for reason when waiving fees
        if op == 'waive' and reason is None:
            reason = 'OTHER'

        r = self._api_call('post',
                           f'{self.api_base_url_users}/{self.user.primary_id}/fees/{self.fee_id}',
                           headers=self._get_headers(),
                           data=None,
                           params={'op': op,
                                   'amount': str(amount),
                                   'comment': comment,
                                   'method': method,
                                   'reason': reason,
                                   'external_transaction_id': external_transaction_id})
        if r.ok is True:
            self.data = JsonData(r.json())
            logging.info(f'{repr(self)}: fee operation "{op}" succeed')
        else:
            self._handle_error(r, f'{repr(self)}: fee operation "{op}" failed')

        return self


if __name__ == "__main__":
    pass



"""This module allows to get and update information about collections"""

from typing import Optional, Literal, List, ClassVar, Union
import logging
from ..record import Record, check_error, JsonData
import almapiwrapper.inventory as inventory
import json

import os


class Collection(Record):
    """
    Class representing a collection object

    Collections can be in NZ and in IZ. They contain list of bibliographic records.

    :ivar zone: zone of the record
    :ivar env: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar data: :class:`almapiwrapper.record.JsonData` object or dict,
    useful to force update a record from a backup
    """

    api_base_url_bibs: ClassVar[str] = f'{Record.api_base_url}/bibs'

    def __init__(self,
                 pid: str,
                 zone: str,
                 env: Optional[Literal['P', 'S']] = 'P') -> None:
        """
        Construct a Collection record.

        """
        self.pid = pid
        self._bibs = None
        self.error = False
        self.error_msg = None
        self._data = None
        self.area = 'Bibs'
        self.format = 'json'
        self.zone = zone
        self.env = env

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.pid}', '{self.zone}', '{self.env}')"

    def _fetch_data(self) -> Optional[JsonData]:
        """Fetch holding data via API. Store the data in the 'data' attribute.

        :return: None
        """
        r = self._api_call('get',
                           f'{self.api_base_url_bibs}/collections/{self.pid}',
                           headers=self._get_headers())

        if r.ok is True:
            logging.info(f'{repr(self)}: collection data available')
            return JsonData(r.json())
        else:
            self._handle_error(r, 'unable to fetch collection data')



    @check_error
    def save(self) -> 'Collection':
        """save(self) -> 'Collection'
        Save collection in a folder.
        Example: records/UBS_9963486250105504/hol_22314215780005504_01.xml

        :return: Collection
        """
        filepath = f'records/{self.zone}_collections/hol_{self.pid}.xml'
        self._save_from_path(filepath)
        return self

    # @check_error
    # def update(self) -> 'Holding':
    #     """update(self) -> 'Holding'
    #     Update data of a holding.
    #
    #     :return: Holding
    #     """
    #     r = self._api_call('put',
    #                        f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings/{self.holding_id}',
    #                        data=etree.tostring(self.data),
    #                        headers=self._get_headers())
    #
    #     if r.ok is True:
    #         self.data = XmlData(r.content)
    #         logging.info(f'{repr(self)}: holding data updated')
    #     else:
    #         self._handle_error(r, 'unable to update holding data')
    #
    #     return self
    #
    # @check_error
    # def delete(self, force: Optional[bool] = False) -> None:
    #     """delete(self, force: Optional[bool] = False) -> None
    #     Delete holding
    #
    #     :return: None
    #     """
    #     if force is True:
    #         self.delete_items()
    #
    #     r = self._api_call('delete',
    #                        f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings/{self.holding_id}',
    #                        headers=self._get_headers())
    #     if r.ok is True:
    #         logging.info(f'{repr(self)} deleted')
    #     else:
    #         self._handle_error(r, 'unable to delete the holding')
    #
    # @check_error
    # def get_items(self) -> List['inventory.Item']:
    #     """get_items(self) -> List['inventory.Item']
    #     This method is used to retrieve the list of items and loads
    #     the xml data of these items. To avoid reloading this information,
    #     the items references are stored in the private attribute '_items'.
    #
    #     :return: list of :class:`almapiwrapper.bib.Item` objects
    #     """
    #     # Check if items have been already fetched
    #     if self._items is not None:
    #         return self._items
    #
    #     # Fetch the item's data through apis
    #     r = self._api_call('get',
    #                        f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings/{self.holding_id}/items',
    #                        params={'limit': '100'},
    #                        headers=self._get_headers())
    #
    #     if r.ok is True:
    #         root = etree.fromstring(r.content, parser=self.parser)
    #     else:
    #         self._handle_error(r, 'unable to fetch items')
    #         return []
    #
    #     items_data = root.findall('.//item_data')
    #     if len(items_data) > 99:
    #         logging.warning(f'{repr(self)}: at least 100 items, risk of missing items')
    #
    #     # No item found
    #     if len(items_data) == 0:
    #         logging.warning(f'{repr(self)}: no item found')
    #         self._items = []
    #         return self._items
    #
    #     logging.info(f'{repr(self)}: {len(items_data)} items fetched')
    #
    #     # Fetch items one by one
    #     self._items = []
    #     for item in items_data:
    #         item_id = item.find('.//pid').text
    #         self._items.append(inventory.Item(holding=self, item_id=item_id))
    #
    #     return self._items
    #
    # @check_error
    # def delete_items(self) -> None:
    #     """delete_items(self) -> None
    #     Delete all items of the holding
    #
    #     :return: None
    #     """
    #     for item in self.get_items():
    #         item.delete()
    #
    # @property

    def _fetch_bibs(self):
        """Fetch bibs of the collection via API. Store the data in the 'bibs' attribute.

        :return: None
        """

        self._bibs = []
        rec_count = None
        mms_ids = []
        while rec_count is None or len(mms_ids) < rec_count:
            r = self._api_call('get',
                               f'{self.api_base_url_bibs}/collections/{self.pid}/bibs',
                               params={'limit': '100', 'offset': str(len(mms_ids))},
                               headers=self._get_headers())
            if r.ok is False:
                self._handle_error(r, f'{repr(self)}: unable to fetch set members')
                return
            data = r.json()
            rec_count = data['total_record_count']
            if 'bib' in data:
                mms_ids += [rec['mms_id'] for rec in data['bib']]
                logging.info(f'{repr(self)}: {len(mms_ids)} / {rec_count} records fetched')
        if self.zone != 'NZ':
            self._bibs = [inventory.IzBib(mms_id, self.zone, self.env) for mms_id in mms_ids]
        else:
            self._bibs = [inventory.NzBib(mms_id, self.env) for mms_id in mms_ids]


    @property
    @check_error
    def bibs(self) -> Optional[List]:
        """bibs(self) -> Optional[List]
        Property of the collection returning the list containing bib records

        :return: List containing bib records
        """
        if self._bibs is None:
            self._fetch_bibs()

        return self._bibs

    def add_bib(self, bib: Union[inventory.IzBib, inventory.NzBib, str]) -> None:
        """add_bib(self, bib: Union[inventory.IzBib, inventory.NzBib]) -> None
        Add a bib to the collection

        :param bib: bib to add to the collection

        :return: None
        """

        if isinstance(bib, inventory.IzBib) or isinstance(bib, inventory.NzBib):
            mms_id = bib.mms_id
        else:
            mms_id = bib
        xml = f'<bib><mms_id>{mms_id}</mms_id></bib>'
        # jsondata = {'bib':[{'mms_id': mms_id}] }
        r = self._api_call('post',
                            f'{self.api_base_url_bibs}/collections/{self.pid}/bibs',
                            data=xml,
                            headers=self._get_headers(data_format='xml'))

        if r.ok is False:
            self._handle_error(r, f'{repr(self)}: unable to add bib {mms_id} to collection')
            return
        self._bibs = None
        logging.info(f'{repr(self)}: bib {mms_id} added to collection')

    def remove_bib(self, bib: Union[inventory.IzBib, inventory.NzBib, str]) -> None:
        """remove_bib(self, bib: Union[inventory.IzBib, inventory.NzBib]) -> None
        Remove a bib from the collection

        :param bib: bib to remove from the collection

        :return: None
        """

        if isinstance(bib, inventory.IzBib) or isinstance(bib, inventory.NzBib):
            mms_id = bib.mms_id
        else:
            mms_id = bib

        r = self._api_call('delete',
                            f'{self.api_base_url_bibs}/collections/{self.pid}/bibs/{mms_id}',
                            headers=self._get_headers())

        if r.ok is False:
            self._handle_error(r, f'{repr(self)}: unable to remove bib {mms_id} from collection')
            return
        self._bibs = None
        logging.info(f'{repr(self)}: bib {mms_id} removed from collection')



    # @library.setter
    # @check_error
    # def library(self, library_code: str) -> None:
    #     """library(self, library_code: str) -> None
    #     This setter is able to update the 852$b of the holding. But the field should already exist.
    #
    #     :param library_code: code of the library to insert in 852$b field
    #
    #     :return: None
    #     """
    #     library = self.data.find('.//datafield[@tag="852"]/subfield[@code="b"]')
    #     if library is None:
    #         logging.error(f'{repr(self)}: no library field in the holding -> not possible to update it')
    #         self.error = True
    #     else:
    #         logging.info(f'{repr(self)}: library changed from "{library.text}" to "{library_code}"')
    #         library.text = library_code
    #
    # @property
    # @check_error
    # def location(self) -> Optional[str]:
    #     """location(self) -> Optional[str]
    #     Property of the holding returning the library code
    #
    #     :return: str containing library code
    #     """
    #     location = self.data.find('.//datafield[@tag="852"]/subfield[@code="c"]')
    #     if location is None:
    #         logging.warning(f'{repr(self)}: no location in the holding')
    #         return
    #     return location.text
    #
    # @location.setter
    # @check_error
    # def location(self, location_code: str) -> None:
    #     """location(self, location_code: str) -> None
    #     This setter is able to update the 852$c of the holding. But the field should already exist.
    #
    #     :param location_code: location code to set to the holding
    #
    #     :return: None
    #     """
    #     location = self.data.find('.//datafield[@tag="852"]/subfield[@code="c"]')
    #     if location is None:
    #         logging.error(f'{repr(self)}: no location field in the holding -> not possible to update it')
    #     else:
    #         logging.info(f'{repr(self)}: location changed from "{location.text}" to "{location_code}"')
    #         location.text = location_code
    #
    # @property
    # @check_error
    # def callnumber(self) -> Optional[str]:
    #     """callnumber(self) -> Optional[str]
    #     Property of the holding returning the callnumber
    #
    #     :return: str containing callnumber
    #     """
    #     field852 = self.data.find('.//datafield[@tag="852"]')
    #     if field852 is None:
    #         return None
    #
    #     callnumber_field = field852.find('./subfield[@code="j"]')
    #
    #     if callnumber_field is None:
    #         callnumber_field = field852.find('./subfield[@code="h"]')
    #
    #     if callnumber_field is None:
    #         logging.warning(f'{repr(self)}: no callnumber field in the holding')
    #         return None
    #
    #     return callnumber_field.text
    #
    # @callnumber.setter
    # @check_error
    # def callnumber(self, callnumber_txt: str) -> None:
    #     """callnumber(self, callnumber_txt: str) -> None
    #     This setter is able to update the 852$j or 852$h of the holding. But the field should already exist.
    #
    #     :param callnumber_txt: text of the callnumber to be set in j or h subfield
    #
    #     :return: None
    #     """
    #     field852 = self.data.find('.//datafield[@tag="852"]')
    #     if field852 is None:
    #         return
    #
    #     callnumber_field = field852.find('./subfield[@code="j"]')
    #
    #     if callnumber_field is None:
    #         callnumber_field = field852.find('./subfield[@code="h"]')
    #
    #     if callnumber_field is None:
    #         logging.error(f'{repr(self)}: no callnumber field in the holding -> not possible to update it')
    #         return
    #
    #     logging.info(f'{repr(self)}: callnumber changed from "{callnumber_field.text}" to "{callnumber_txt}"')
    #     callnumber_field.text = callnumber_txt
    #
    # @staticmethod
    # def get_data_from_disk(mms_id: str, holding_id: str, zone: str) -> Optional[XmlData]:
    #     """get_data_from_disk(mms_id, holding_id, zone)
    #     Fetch the data of the described record
    #
    #     :param mms_id: bib record mms_id
    #     :param holding_id: holding record ID
    #     :param zone: zone of the record
    #
    #     :return: :class:`almapiwrapper.record.XmlData` or None
    #     """
    #     if os.path.isdir(f'records/{zone}_{mms_id}') is False:
    #         return
    #
    #     # Fetch all available filenames related to this record
    #     file_names = sorted([file_name for file_name in os.listdir(f'records/{zone}_{mms_id}')
    #                          if file_name.startswith(f'hol_{holding_id}') is True])
    #
    #     if len(file_names) == 0:
    #         return
    #
    #     return XmlData(filepath=f'records/{zone}_{mms_id}/{file_names[-1]}')



from .recset import RecSet, LogicalSet, ItemizedSet, NewLogicalSet, NewItemizedSet
from .job import Job
from .reminder import Reminder, fetch_reminders
from .library import Library, fetch_libraries, Location, OpenHours



from ..record import Record, check_error
from typing import Optional, ClassVar, Literal
import almapiwrapper.analytics as analyticslib
from lxml import etree
import pandas as pd
import logging


class AnalyticsReport(Record):
    """Class representing an Analytics report

    :cvar api_base_url_analytics: url of the analytics api
    :ivar path: path of the report
    :ivar zone: initial value: zone of the report
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar data: :class:`almapiwrapper.record.XmlData` with raw report data
    :ivar report_name: name of the report
    :ivar filter: filter of the report
    """

    api_base_url_analytics: ClassVar[str] = 'https://api-eu.hosted.exlibrisgroup.com/almaws/v1/analytics/reports'

    def __init__(self,
                 path: str,
                 zone: str,
                 env: Literal['P', 'S'] = 'P',
                 filter_to_apply: Optional[str] = None) -> None:
        """Constructor of AnalyticsReport Object
        """

        super().__init__(zone, env)
        self.area = 'Analytics'
        self.format = 'xml'
        self.path = path
        self.report_name = self.path.split('/')[-1]
        self.filter = filter_to_apply

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.
        :return: str
        """
        return f"{self.__class__.__name__}('{self.path}', '{self.zone}', '{self.env}')"


    def _fetch_data(self) -> Optional[pd.DataFrame]:
        """Fetch the json data of the AnalyticsReport

        :return: pandas DataFrame"""

        # Construct params
        params = {"path": self.path,
                  "limit": 1000}
        if self.filter is not None:
            params['filter'] = self.filter

        r = self._api_call('get',
                           f'{self.api_base_url_analytics}',
                           params=params,
                           headers=self._get_headers(rights='R'))
        if r.ok is False:
            self._handle_error(r, f'{repr(self)}: unable to fetch Analytics data')
            return None

        logging.info(f'{repr(self)}: Analytics data available')

        # Resumption token are unique for analytics reports, they are the same for all queries
        resumption_field = etree.fromstring(r.content).find('.//ResumptionToken')
        if resumption_field is not None:
            params = {'limit': '1000',
                      'token': resumption_field.text}

            final_columns = [col.get('{urn:saw-sql}columnHeading') for col in
                             etree.fromstring(r.content).findall('.//{http://www.w3.org/2001/XMLSchema}element')]

            temp_columns = [f'Column{n}' for n in range(len(final_columns))]
            df = pd.DataFrame(columns=temp_columns)

            while True:
                for row in etree.fromstring(r.content).findall('.//{urn:schemas-microsoft-com:xml-analysis:rowset}Row'):
                    data = {}

                    for cell in row:
                        data[cell.tag[cell.tag.find('Column'):]] = cell.text
                    df.loc[len(df)] = data

                logging.info (f'{repr(self)}: {len(df)} rows fetched')

                if etree.fromstring(r.content).find('.//IsFinished').text == 'true':
                    break

                r = self._api_call('get',
                                   f'{self.api_base_url_analytics}',
                                   params=params,
                                   headers=self._get_headers(rights='R'))

                if r.ok is False:
                    self._handle_error(r, f'{repr(self)}: unable to fetch Analytics data')
                    return None

            df.columns = final_columns
            df.drop('0', axis=1, inplace=True)
            return df

    @check_error
    def save(self, format: Optional[Literal['json', 'csv']] = 'csv') -> 'analyticslib.AnalyticsReport':
        """save(self, format: Optional[Literal['json', 'csv']] = 'csv') -> 'analyticslib.AnalyticsReport'
        Save a user fee record in the 'records' folder

        When saved, a suffix is added to the file path with the version.
        Example: records/report_<report_name>/report_<report_name>_<IZ>_<version>.xml

        :return: object :class:`almapiwrapper.analytics.AnalyticsReport`
        """
        filepath = f'records/report_{self.report_name}/report_{self.report_name}_{self.zone}.{format}'
        self._save_from_path(filepath)
        return self

if __name__ == "__main__":
    pass


import unittest
from dateutil import parser
from datetime import timedelta
from almapiwrapper.users import User, NewUser, Loan, Request
from almapiwrapper.record import JsonData
from almapiwrapper.inventory import Item
from almapiwrapper import config_log
import os

config_log("test.log")

if os.getcwd().endswith('test'):
    os.chdir('..')

class TestRequest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        for primary_id in ['testRequestUser1',]:
            u = User(primary_id, 'UBS', 'S')

            if u.data is None:

                # Create new user
                data = JsonData(filepath=f'test/data/user_{primary_id}.json')

                user = NewUser('UBS', 'S', data).create()


        # Item('996259850105504', '22202377380005504', '23202377370005504', 'UBS', 'S').scan_in('A100', 'A100_KUR')
        # Item('9953896770105504', '22315547880005504', '23315547870005504', 'UBS', 'S').scan_in('A100', 'A100_KUR')

    def test_fetch_request(self):

        u = User('testRequestUser1', 'UBS', 'S')

        # Create new request
        req_data = JsonData(filepath=f'test/data/request_test1.json')
        _ = Request(zone='UBS', env='S', data=req_data).create()
        r = u.requests[0]
        req_id = r.request_id
        self.assertFalse(r.error, 'Error during request creation')

        self.assertEqual(r.data['title'],
                         'The dot-com debacle and the return of reason Louis E.V. Nevaer',
                         'Request title mismatch')
        r2 = Request(request_id=req_id, zone='UBS', env='S')
        _ = r2.data
        self.assertFalse(r2.error, 'Error during request fetch')
        self.assertEqual(r2.data['title'],
                         'The dot-com debacle and the return of reason Louis E.V. Nevaer',
                         'Request title mismatch')
        self.assertEqual(r2.user.primary_id, 'testRequestUser1', 'Request user mismatch')

    def test_create_request(self):
        req_data = JsonData(filepath=f'test/data/request_test2.json')
        _ = Request(zone='UBS', env='S', data=req_data).create()
        reqs = User('testRequestUser1', 'UBS', 'S').requests
        self.assertGreater(len(reqs), 0, 'No request found')
        for reqs in reqs:
            self.assertIn(reqs.data['title'],
                          ['The dot-com debacle and the return of reason Louis E.V. Nevaer',
                           'Marketing data science modeling techniques in predictive analytics with R and Python Thomas W. Miller'],
                          'Request title mismatch')

    def test_cancel_request(self):
        req_data = JsonData(filepath=f'test/data/request_test3.json')
        r = Request(zone='UBS', env='S', data=req_data).create()
        req_id = r.data['request_id']
        r.cancel()
        r = Request(request_id=req_id, user_id='testRequestUser1' , zone='UBS', env='S')
        self.assertEqual(r.data['request_status'], 'HISTORY', 'Request status mismatch')
        # print(r)


    @classmethod
    def tearDownClass(cls):
        for primary_id in ['testRequestUser1']:
            u = User(primary_id, 'UBS', 'S')
            u.delete()

if __name__ == '__main__':
    unittest.main()



"""This module allow to create and manage sets in Alma"""
from typing import Literal, Optional, Dict, Union
import logging
from ..record import XmlData, JsonData, Record, check_error


class Job(Record):
    """Class representing a job
    :param job_id: ID of the job
    :param zone: zone of the job
    :param env: environment of the entity: 'P' for production and 'S' for sandbox
    :param job_type: type of the job: 'M' for manual, 'S' for scheduled

    :ivar job_id: initial value: ID of the job
    :ivar zone: initial value: zone of the fee
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar job_type: initial value: type of the job: 'M' for manual, 'S' for scheduled
    :ivar instance_id: initial value: str
    :ivar area: initial value: 'Conf'
    :ivar format: initial value: 'xml'
    """
    def __init__(self,
                 job_id: str,
                 zone: str,
                 env: Literal['P', 'S'] = 'P',
                 job_type: Literal['M', 'S', 'O'] = 'M') -> None:
        """Constructor of `Job`
        """
        super().__init__(zone, env)
        self.area = 'Conf'
        self.format = 'xml'
        self.job_id = job_id
        self.job_type = job_type
        self.instance_id = None

    def _fetch_data(self) -> Optional[XmlData]:
        """This method fetch the data describing the job.

        :return: :class:`almapiwrapper.record.XmlData`
        """
        r = self._api_call('get',
                           f'{self.api_base_url}/conf/jobs/{self.job_type}{self.job_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: job data available')
            return XmlData(r.content)
        else:
            self._handle_error(r, 'unable to fetch job data')

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.job_id}', '{self.zone}', '{self.env}')"

    @check_error
    def run(self, parameters: Optional[Union[str, 'XmlData']] = '<job/>') -> Optional['XmlData']:
        """run(parameters: Optional[Union[str, 'XmlData']] = '<job/>') -> Optional['XmlData']
        Run the job with the provided parameters

        :param parameters: :class:`almapiwrapper.record.XmlData` or string with the parameters to use for the job.
            For some jobs no parameters are required.

        :return: :class:`almapiwrapper.record.XmlData` with the instance of the job or None in case of error.

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """

        # Transform the XmlData parameters to string, useful to make the api request
        if type(parameters) == XmlData:
            parameters = str(parameters)

        r = self._api_call('post',
                           f'{self.api_base_url}/conf/jobs/{self.job_type}{self.job_id}',
                           params={'op': 'run'},
                           data=parameters,
                           headers=self._get_headers())

        if r.ok is True:
            result = XmlData(r.content)

            # Fetch the instance ID
            link = result.content.find('.//additional_info').get('link')
            instance_id = link.split('/')[-1]
            logging.info(f'{repr(self)}: run job successful, instance: {instance_id}')

            # Store the current instance ID in the instance_id attribute of the job
            self.instance_id = instance_id
            return result
        else:
            self._handle_error(r, 'unable to run the job')

    @check_error
    def get_instance_info(self, instance_id: Optional[str] = None) -> Optional['JsonData']:
        """get_instance_info(instance_id: Optional[str] = None) -> Optional['JsonData']
        Get information about a specif job instance

        :param instance_id: optional parameter indicating the instance to check. If not
            provided it will use the instance id of a job started

        :return: :class:`almapiwrapper.record.JsonData` with the details of a job instance
            return None in case of error

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        if instance_id is None:
            instance_id = self.instance_id

        if instance_id is None:
            logging.error(f'{repr(self)}: No instance job ID available or provided')
            return

        r = self._api_call('get',
                           f'{self.api_base_url}/conf/jobs/{self.job_type}{self.job_id}/instances/{instance_id}',
                           headers=self._get_headers(data_format='json'))
        if r.ok is True:
            logging.info(f'{repr(self)}: job information available for instance {instance_id}')
            return JsonData(r.json())
        else:
            self._handle_error(r, 'unable to fetch job data')

    @check_error
    def check_instance_state(self, instance_id: Optional[str] = None) -> Optional[Dict]:
        """check_instance_state(instance_id: Optional[str] = None) -> Optional[Dict]
        Check the state of the provided instance id or of instance started
        with :meth:`almapiwrapper.config.Job.run`

        :param instance_id: optional parameter indicating the instance to check. If not
            provided it will use the instance id of a job started

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        data = self.get_instance_info(instance_id)

        if data is not None:
            return {'status': data.content['status']['value'],
                    'progress': data.content['progress']}

    @check_error
    def get_instances(self) -> Optional['JsonData']:
        """get_instances(self) -> Optional['JsonData']
        Get the list of the instances of a job in json format

        :return: :class:`almapiwrapper.record.JsonData` with the list of job instances
            return None in case of error

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        r = self._api_call('get',
                           f'{self.api_base_url}/conf/jobs/{self.job_type}{self.job_id}/instances',
                           headers=self._get_headers(data_format='json'))
        if r.ok is True:
            logging.info(f'{repr(self)}: job information available for instances')
            return JsonData(r.json())
        else:
            self._handle_error(r, 'unable to fetch job data')
